// copyright 2012–2021 nick black
#ifndef GROWLIGHT_SMART
#define GROWLIGHT_SMART

#ifdef __cplusplus
extern "C" {
#endif

struct device;

int probe_smart(struct device *d);

#ifdef __cplusplus
}
#endif

#endif
